/******************************************************************************
*
* project name:    TICHESS
* project version: 2.01
* file name:       hash.c
* initial date:    10/03/2000
* author:          thomas.nussbaumer@peem.com
* description:     contains the hashing stuff
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.04: 22/03/2000
*        usage of define DEVELOPER_VERSION integrated (exclusion of special
*        functions)
*        bug fixed in XorHKey(): have to add additionally 6 to "negative" figs
*
* v1.03: 17/03/2000
*        better hashing strategy added (can be turned off with define)
*
* v1.02: 16/03/2000
*        problem with king fixed again ;-)
*        hashing is done now absolutely simple 
*
* v1.01: 11/03/2000
*        problem "king not moved out of check" fixed by always hashing checks
*        hashcodes array will be allocated now
*
* v1.00: 10/03/2000
*        initial version
*
******************************************************************************/

//=============================================================================
//
// The following define controls the style of hashing. If it is set to 1
// simple hashing is used, where each key maps only to one entry and replaces
// it.
//
// If is set to 0 a more sophisticated mapping is used where each key maps
// to 2 entries: it's normal entry and the entry which is resolved if each bit
// of the key is inverted.
// If both entries matches to the same position the best of them will be
// restored. If storing a move the worst of both entries gets overwritten.
//
//=============================================================================

#define SIMPLE_HASH 0

hash_t act_hashkey = 0;

extern hash_t**     hashcodes;   
extern hentry_t*    hashtable_w;
extern hentry_t*    hashtable_b;
extern char*        board;
extern int          act_depth;
extern unsigned int move_count;
extern int          act_color;
extern int          act_color;
extern int          chess_indicator;


int           hashmode  = TRUE;  // state of hashing usage
int           hashprint = FALSE; // state of hash statistics
unsigned long used_hash = 0;     
unsigned long hash_sum  = 0;
unsigned long node_sum  = 0;

char          default_hashmode = 0;

/*===========================================================================*/
/* is used to toggle hash table usage                                        */
/*===========================================================================*/
void ResetHashCounters(void) {
    used_hash = 0;
    hash_sum  = 0;
    node_sum  = 0;
}


#if DEVELOPER_VERSION
/*===========================================================================*/
/* a helper function to output the 64bit hashcode                            */
/*===========================================================================*/
char szhash[20] = {0};

char* Hash2Str(hash_t code) {
    unsigned long hi;
    unsigned long lo;

    lo = code & 0xffffffff;
    hi = code >> 32;
    
    sprintf(szhash,"%08lX%08lX\n",hi,lo);
    return szhash;
}


/*===========================================================================*/
/* is used to toggle hash table usage                                        */
/*===========================================================================*/
inline void ToggleHashMode(void) {
    if (default_hashmode == OPTION_OFF) return;
    if (hashmode) {
        hashmode = FALSE;
        puts("hashing now OFF\n");
    }
    else {
       hashmode = TRUE;
       puts("hashing now ON\n");
    }  
}   


/*===========================================================================*/
/* is used to toggle hash statistics printing                                */
/*===========================================================================*/
inline void ToggleHashStatistics(void) {
    if (default_hashmode == OPTION_OFF) return;
    if (hashprint) {
        hashprint = FALSE;
        puts("statistics now OFF\n");
    }
    else {
        hashprint = TRUE;
        puts("statistics now ON\n");
    }  
}   
#endif


/*===========================================================================*/
/* show nodes statistics if hashprint == TRUE                                */
/*                                                                           */
/* SIDE EFFECT: counters will be increased even if hashprint == FALSE        */
/*===========================================================================*/
inline void ShowNodes2Hashed(unsigned long nodes,int alpha) {
    if (hashmode) {
        hash_sum += used_hash;
    }
    node_sum += nodes;
    if (hashprint) {
#if TI92_VERSION
    printf("%lu/%lu:%lu/%lu(%d)\n",used_hash,nodes,hash_sum,node_sum,alpha);
#else
    printf("%lu/%lu(%d)\n%lu/%lu\n",used_hash,nodes,alpha,hash_sum,node_sum,alpha);
#endif
    }
    used_hash = 0;
}


/*===========================================================================*/
/* try to retrieve a value for this position from hash                       */
/* returns TRUE if value could be retrieved                                  */
/*===========================================================================*/
#if SIMPLE_HASH
inline int RetrieveFromHash(int *value) {
    unsigned int  key;
    hentry_t*     entry;

    if (!hashmode) return FALSE;

    key = (unsigned int)(act_hashkey & (HASHSIZE-1));

    if (act_color == WHITE) entry = &hashtable_w[key];
    else                    entry = &hashtable_b[key];

    if (entry->key != act_hashkey) return FALSE;
    
    used_hash++;
    *value = entry->val;
    
    return TRUE;
}
#else
inline int RetrieveFromHash(int *value) {
    unsigned int  key1;
    unsigned int  key2;
    hentry_t*     entry1;
    hentry_t*     entry2;

    if (!hashmode) return FALSE;

    key1 = (unsigned int)(act_hashkey & (HASHSIZE-1));
    key2 = ~key1 & (HASHSIZE-1);

    if (act_color == WHITE) {
        entry1 = &hashtable_w[key1];

        if (entry1->key != act_hashkey) {
            entry2 = &hashtable_w[key2];
            if (entry2->key != act_hashkey) return FALSE;
            used_hash++;
            *value = entry2->val;
            return TRUE;
        }
        else {                
            entry2 = &hashtable_w[key2];
            if (entry2->key != act_hashkey) {                
                used_hash++;
                *value = entry1->val;
                return TRUE;
            }
            if (entry2->val > entry1->val) {
                *value = entry2->val;
            }
            else {
                *value = entry1->val;
            }
            used_hash++;
            return TRUE;
        }
    }
    else {
        entry1 = &hashtable_b[key1];

        if (entry1->key != act_hashkey) {
            entry2 = &hashtable_b[key2];
            if (entry2->key != act_hashkey) return FALSE;
            used_hash++;
            *value = entry2->val;
            return TRUE;
        }
        else {                
            entry2 = &hashtable_b[key2];
            if (entry2->key != act_hashkey) {                
                used_hash++;
                *value = entry1->val;
                return TRUE;
            }
            if (entry2->val < entry1->val) {
                *value = entry2->val;
            }
            else {
                *value = entry1->val;
            }
            used_hash++;
            return TRUE;
        }
    }
}
#endif

/*===========================================================================*/
/* stores a value in the hash                                                */
/*===========================================================================*/
#if SIMPLE_HASH
inline void StoreInHash(int value) {
    unsigned int  key;
    hentry_t*     entry;

    if (!hashmode) return;

    key = (unsigned int)(act_hashkey & (HASHSIZE-1));

    if (act_color == WHITE) entry = &hashtable_w[key];
    else                    entry = &hashtable_b[key];
    
    entry->key = act_hashkey;
    entry->val = value;
}
#else 
inline void StoreInHash(int value) {
    unsigned int  key1;
    unsigned int  key2;
    hentry_t*     entry1;
    hentry_t*     entry2;

    if (!hashmode) return;

    key1 = (unsigned int)(act_hashkey & (HASHSIZE-1));
    key2 = ~key1 & (HASHSIZE-1);

    if (act_color == WHITE) {
        entry1 = &hashtable_w[key1];
        entry2 = &hashtable_w[key2];
        
        if (entry1->val < entry2->val) {
            entry1->key = act_hashkey;
            entry1->val = value;
        }
        else {
            entry2->key = act_hashkey;
            entry2->val = value;
        }
    }
    else {
        entry1 = &hashtable_b[key1];
        entry2 = &hashtable_b[key2];
        
        if (entry1->val > entry2->val) {
            entry1->key = act_hashkey;
            entry1->val = value;
        }
        else {
            entry2->key = act_hashkey;
            entry2->val = value;
        }
    }
}
#endif


int init_random = 1;


unsigned long x[55] = {
       1410651636UL, 3012776752UL, 3497475623UL, 2892145026UL, 1571949714UL,
       3253082284UL, 3489895018UL,  387949491UL, 2597396737UL, 1981903553UL,
       3160251843UL,  129444464UL, 1851443344UL, 4156445905UL,  224604922UL,
       1455067070UL, 3953493484UL, 1460937157UL, 2528362617UL,  317430674UL, 
       3229354360UL,  117491133UL,  832845075UL, 1961600170UL, 1321557429UL,
        747750121UL,  545747446UL,  810476036UL,  503334515UL, 4088144633UL,
       2824216555UL, 3738252341UL, 3493754131UL, 3672533954UL,   29494241UL,
       1180928407UL, 4213624418UL,   33062851UL, 3221315737UL, 1145213552UL,
       2957984897UL, 4078668503UL, 2262661702UL,   65478801UL, 2527208841UL,
       1960622036UL,  315685891UL, 1196037864UL,  804614524UL, 1421733266UL,
       2017105031UL, 3882325900UL,  810735053UL,  384606609UL, 2393861397UL 
};

unsigned long y[55] = {};
int           jjj   = 0;
int           kkk   = 0;

/*===========================================================================*/
/* 32 bit random number generator                                            */
/*===========================================================================*/
unsigned long Random32(void) {
    unsigned long  ul;
  
    if (init_random) {
       int i;
    
       init_random = 0;
       for (i = 0; i < 55; i++) y[i] = x[i];
       jjj = 24 - 1;
       kkk = 55 - 1;
    }
  
    ul = (y[kkk] += y[jjj]);
    if (--jjj < 0) jjj = 55 - 1;
    if (--kkk < 0) kkk = 55 - 1;
    return(ul);
}


/*===========================================================================*/
/* setup the hashcode table by assigning each square/piece/side combination  */
/* an own 64bit hashcode                                                     */
/*===========================================================================*/
inline void InitHashCode() {
    int    figure, sq, i;
    hash_t result;

    if (default_hashmode == OPTION_OFF) {
        hashmode  = FALSE;
        hashprint = FALSE;
    }
    else {
       ResetHashCounters();
       hashmode  = TRUE;
       hashprint = FALSE;
       memset(hashtable_w,0,sizeof(hentry_t)*HASHSIZE);
       memset(hashtable_b,0,sizeof(hentry_t)*HASHSIZE);
    }
    init_random = 1;


    for (figure = 0; figure < 12; figure++) {
        for (sq = 0; sq < 64; sq++) {
            result = Random32();
            result <<= 32;
            result |= Random32();
	    hashcodes[figure][sq] = result;
	}
    }
}


/*===========================================================================*/
/* setup the hashcode table by assigning each square/piece/side combination  */
/* an own 64bit hashcode                                                     */
/*===========================================================================*/
inline void XorHKey(char fig,int square) {
    if (fig==EMPTY || fig==OUTSIDE) return;

    if (fig < 0) fig = -fig + 6;
    fig--;

    act_hashkey ^= hashcodes[fig][(square/10 - 2) * 8 + (square%10 - 1)];
}


/*===========================================================================*/
/* the hashcode is generated by looping through all squares and XORing all   */
/* hashcodes of the square/piece/color combinations found                    */
/*===========================================================================*/
void CalcHashKey() {
    int  sq;

    act_hashkey = 0;
    for (sq=A1; sq<=H8; sq++) XorHKey(board[sq],sq);
}


